import json
import unittest
import six
if six.PY3:
    from unittest.mock import Mock, patch
else:
    from mock import Mock, patch

from ..exceptions import PapermillException
from ..iorw import HttpHandler


class TestHttpHandler(unittest.TestCase):
    """
    Tests for `HttpHandler`.
    """

    def test_listdir(self):
        """
        Tests that listdir raises `PapermillException` saying that that
        `listdir` function is not supported.
        """
        with self.assertRaises(PapermillException) as e:
            HttpHandler.listdir('http://example.com')

        self.assertEqual(
                '{}'.format(e.exception), 'listdir is not supported by HttpHandler'
        )

    def test_read(self):
        """
        Tests that the `read` function performs a request to the giving path
        and returns the response.
        """
        path = 'http://example.com'
        text = 'request test response'

        with patch('papermill.iorw.requests.get') as mock_get:
            mock_get.return_value = Mock(text=text)
            self.assertEqual(HttpHandler.read(path), text)
            mock_get.assert_called_once_with(path)

    def test_write(self):
        """
        Tests that the `write` function performs a request a put request to
        the giving path.
        """
        path = 'http://example.com'
        buf = '{"papermill": true}'

        with patch('papermill.iorw.requests.put') as mock_put:
            HttpHandler.write(buf, path)
            mock_put.assert_called_once_with(path, json=json.loads(buf))



# -*- coding: utf-8 -*-
"""
api.py

Provides the API for papermill

"""
from __future__ import unicode_literals
import os

import IPython
from IPython.display import display as ip_display, Markdown
import pandas as pd
from six import string_types

from .exceptions import PapermillException
from .iorw import load_notebook_node, list_notebook_files

RECORD_OUTPUT_TYPE = 'application/papermill.record+json'
DISPLAY_OUTPUT_TYPE = 'application/papermill.display+json'


def record(name, value):
    """
    Record a value in the output notebook when a cell is executed.

    The recorded value can be retrieved during later inspection of the
    output notebook.

    Example
    -------
    `record` provides a handy way for data to be stored with a notebook to
    be used later::

        pm.record("hello", "world")
        pm.record("number", 123)
        pm.record("some_list", [1, 3, 5])
        pm.record("some_dict", {"a": 1, "b": 2})

    pandas can be used later to recover recorded values by
    reading the output notebook into a `dataframe`::

        nb = pm.read_notebook('notebook.ipynb')
        nb.dataframe

    Parameters
    ----------
    name : str
        Name of the value to record.
    value: str, int, float, list, dict
        The value to record.

    """
    # IPython.display.display takes a tuple of objects as first parameter
    # `http://ipython.readthedocs.io/en/stable/api/generated/IPython.display.html#IPython.display.display`
    data = {RECORD_OUTPUT_TYPE: {name: value}}
    ip_display(data, raw=True)


def display(name, obj):
    """
    Display an object with the reference `name`.

    Parameters
    ----------
    name : str
        Name of the output.
    obj : object
        An object that can be displayed in the notebook.

    """
    data, metadata = IPython.core.formatters.format_display_data(obj)
    metadata['papermill'] = dict(name=name)
    ip_display(data, metadata=metadata, raw=True)


def read_notebook(path):
    """
    Returns a Notebook object loaded from the location specified at `path`.

    Parameter
    ---------
    path : str
        Path to a notebook `.ipynb` file.

    Returns
    -------
    notebook : object
        A Notebook object.

    """
    if not path.endswith(".ipynb"):
        raise PapermillException(
            "Requires an '.ipynb' file extension. Provided path: '%s'",
            path)

    nb = Notebook()
    nb.path = path
    nb.node = load_notebook_node(path)
    return nb


def read_notebooks(path):
    """
    Returns a NotebookCollection including the notebooks read from the
    directory specified by `path`.

    Parameter
    ---------
    path : str
        Path to directory containing notebook `.ipynb` files.

    Returns
    -------
    nb_collection : object
        A `NotebookCollection` object.

    """
    nb_collection = NotebookCollection()
    for notebook_path in list_notebook_files(path):
        fn = os.path.basename(notebook_path)
        nb_collection[fn] = read_notebook(notebook_path)
    return nb_collection


class Notebook(object):
    """
    Representation of a notebook.

    Parameters
    ----------
    node : `nbformat.NotebookNode`
        a notebook object
    path : str, optional
        the path to the notebook

    Method
    ------
    display_output


    """
    def __init__(self, node=None, path=None):

        if path is not None and not node:
            raise ValueError('notebook must be defined when path is given')
        self.path = path or ''
        self.node = node

    @property
    def filename(self):
        """str: filename found a the specified path"""
        return os.path.basename(self.path)

    @property
    def directory(self):
        """str: directory name at the specified path"""
        return os.path.dirname(self.path)

    @property
    def version(self):
        """str: version of Jupyter notebook spec"""
        return _get_papermill_metadata(self.node, 'version', default=None)

    @property
    def parameters(self):
        """dict: parameters stored in the notebook metadata"""
        return _get_papermill_metadata(self.node, 'parameters', default={})

    @property
    def environment_variables(self):
        """dict: environment variables used by the notebook execution"""
        return _get_papermill_metadata(
            self.node, 'environment_variables', default={})

    @property
    def data(self):
        """dict: a dictionary of data found in the notebook"""
        return _fetch_notebook_data(self.node)

    @property
    def dataframe(self):
        """pandas dataframe: a dataframe of named parameters and values"""
        df = pd.DataFrame(columns=['name', 'value', 'type', 'filename'])

        # TODO: refactor to a Dict comprehension or list comprehensions
        i = 0
        for name in sorted(self.parameters.keys()):
            df.loc[i] = name, self.parameters[name], 'parameter', self.filename
            i += 1

        for name in sorted(self.data.keys()):
            df.loc[i] = name, self.data[name], 'record', self.filename
            i += 1
        return df

    @property
    def metrics(self):
        """pandas dataframe: dataframe of cell execution counts and times"""
        df = pd.DataFrame(columns=['filename', 'cell', 'value', 'type'])

        for i, cell in enumerate(self.node.cells):
            execution_count = cell.get("execution_count")
            if not execution_count:
                continue
            name = "Out [%s]" % str(execution_count)
            value = cell.metadata.get('papermill', {}).get('duration', 0.0)
            df.loc[i] = self.filename, name, value, "time (s)",
        return df

    def display_output(self, name):
        """Display output from a named source notebook in a running notebook.

        Parameter
        ---------
        name : str
            name of source notebook

        """
        outputs = _get_notebook_outputs(self.node)
        if name not in outputs:
            raise PapermillException(
                "Output Name '%s' is not available in this notebook.")
        output = outputs[name]
        ip_display(output.data, metadata=output.metadata, raw=True)


def _get_papermill_metadata(nb, name, default=None):
    """Returns a dictionary of a notebook's metadata."""
    return nb.metadata.get('papermill', {}).get(name, default)


def _get_notebook_outputs(nb_node):
    """Returns a dictionary of the notebook outputs."""
    outputs = {}
    for cell in nb_node.cells:
        for output in cell.get('outputs', []):
            if 'papermill' in output.get('metadata', {}):
                output_name = output.metadata.papermill.get('name')
                if output_name:
                    outputs[output_name] = output
    return outputs


def _fetch_notebook_data(nb_node):
    """Returns a dictionary of the data recorded in a notebook."""
    notebook_data = {}
    for cell in nb_node.cells:
        for output in cell.get('outputs', []):
            if 'data' in output and RECORD_OUTPUT_TYPE in output['data']:
                notebook_data.update(output['data'][RECORD_OUTPUT_TYPE])
    return notebook_data


class NotebookCollection(object):
    """Represents a collection of notebooks as a dictionary of notebooks.

    """
    def __init__(self):

        self._notebooks = {}

    def __setitem__(self, key, value):
        # If notebook is a path str then load the notebook.
        if isinstance(value, string_types):
            value = read_notebook(value)

        if not isinstance(value, Notebook):
            raise PapermillException(
                "Value must either be a path string or a Papermill Notebook object. Found: '%s'"
                % str(type(value)))
        self._notebooks[key] = value

    def __getitem__(self, key):
        return self._notebooks[key]

    def __delitem__(self, key):
        del self._notebooks[key]

    @property
    def dataframe(self):
        """list: a list of dataframes from a collection of notebooks"""
        df_list = []
        for key in sorted(self._notebooks):
            nb = self._notebooks[key]
            df = nb.dataframe
            df['key'] = key
            df_list.append(df)
        return pd.concat(df_list).reset_index(drop=True)

    @property
    def metrics(self):
        """list: a list of metrics from a collection of notebooks"""
        df_list = []
        for key in sorted(self._notebooks):
            nb = self._notebooks[key]
            df = nb.metrics
            df['key'] = key
            df_list.append(df)
        return pd.concat(df_list).reset_index(drop=True)

    def display_output(self, key, output_name):
        """Display output as markdown.

        Parameters
        ----------
        key : str
            parameter names to be used as markdown headings in output
        output_name : str
            a named notebook to display as output

        """
        if isinstance(key, string_types):
            ip_display(Markdown("### %s" % str(key)))
            self[key].display_output(output_name)
        else:
            for i, k in enumerate(key):
                if i > 0:
                    ip_display(Markdown("<hr>"))  # tag between outputs
                ip_display(Markdown("### %s" % str(k)))
                self[k].display_output(output_name)



import os
import sys

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


def get_notebook_path(*args):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'notebooks', *args)


def get_notebook_dir(*args):
    return os.path.dirname(get_notebook_path(*args))


class RedirectOutput(object):

    def __init__(self):
        self.redirected_stdout = None
        self.redirected_stderr = None
        self.old_stdout = None
        self.old_stderr = None

    def __enter__(self):
        self.old_stdout = sys.stdout
        sys.stdout = self.redirected_stdout = StringIO()
        self.old_stderr = sys.stderr
        sys.stderr = self.redirected_stderr = StringIO()
        return self

    def __exit__(self, *args):
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr

    def get_stdout(self):
        return self.redirected_stdout.getvalue()

    def get_stderr(self):
        return self.redirected_stderr.getvalue()



 # -*- coding: utf-8 -*-

from __future__ import unicode_literals

import datetime
import os
import sys

import nbformat
from concurrent import futures
from jupyter_client.kernelspec import get_kernel_spec
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors.execute import CellExecutionError
from nbconvert.preprocessors.base import Preprocessor
from six import string_types, integer_types
# tqdm creates 2 globals lock which raise OSException if the execution
# environment does not have shared memory for processes, e.g. AWS Lambda
try:
    from tqdm import tqdm
    no_tqdm = False
except OSError:
    no_tqdm = True

from .conf import settings
from .exceptions import PapermillException, PapermillExecutionError
from .iorw import load_notebook_node, write_ipynb, read_yaml_file, get_pretty_path

PENDING = "pending"
RUNNING = "running"
COMPLETED = "completed"


def preprocess(self, nb, resources):
    """
    This function monkey patches the `Preprocessor.preprocess` method.

    We are doing this for the following reasons:

    1. Notebooks will stop executing when they encounter a failure but not
       raise a `CellException`. This allows us to save the notebook with the
       traceback even though a `CellExecutionError` was encountered.

    2. We want to write the notebook as cells are executed. We inject our
       logic for that here.

    3. We want to include timing and execution status information with the
       metadata of each cell.

    Parameters
    ----------
    nb : NotebookNode
        Notebook being converted
    resources : dictionary
        Additional resources used in the conversion process.  Allows
        preprocessors to pass variables into the Jinja engine.

    """
    output_path = nb.metadata.papermill['output_path']

    # Reset the notebook.
    for cell in nb.cells:
        # Reset the cell execution counts.
        if cell.get("execution_count") is not None:
            cell.execution_count = None

        # Clear out the papermill metadata for each cell.
        cell.metadata['papermill'] = dict(
            exception=None,
            start_time=None,
            end_time=None,
            duration=None,
            status=PENDING  # pending, running, completed
        )
        if cell.get("outputs") is not None:
            cell.outputs = []

    # Execute each cell and update the output in real time.
    with futures.ThreadPoolExecutor(max_workers=1) as executor:

        # Generate the iterator
        if self.progress_bar:
            execution_iterator = tqdm(enumerate(nb.cells), total=len(nb.cells))
        else:
            execution_iterator = enumerate(nb.cells)

        for index, cell in execution_iterator:
            cell.metadata["papermill"]["status"] = RUNNING
            future = executor.submit(write_ipynb, nb, output_path)
            t0 = datetime.datetime.utcnow()
            try:
                if not cell.source:
                    continue

                nb.cells[index], resources = self.preprocess_cell(
                    cell, resources, index)
                cell.metadata['papermill']['exception'] = False
                if self.log_output:
                    log_outputs(nb.cells[index])

            except CellExecutionError:
                cell.metadata['papermill']['exception'] = True
                break
            finally:
                t1 = datetime.datetime.utcnow()
                cell.metadata['papermill']['start_time'] = t0.isoformat()
                cell.metadata['papermill']['end_time'] = t1.isoformat()
                cell.metadata['papermill']['duration'] = (
                    t1 - t0).total_seconds()
                cell.metadata['papermill']['status'] = COMPLETED
                future.result()
    return nb, resources


def log_outputs(cell):

    execution_count = cell.get("execution_count")
    if not execution_count:
        return

    stderrs = []
    stdouts = []
    for output in cell.get("outputs", []):
        if output.output_type == "stream":
            if output.name == "stdout":
                stdouts.append("".join(output.text))
            elif output.name == "stderr":
                stderrs.append("".join(output.text))
        elif "data" in output and "text/plain" in output.data:
            stdouts.append(output.data['text/plain'])

    # Log stdouts
    sys.stdout.write('{:-<40}'.format("Out [%s] " % execution_count) + "\n")
    sys.stdout.write("\n".join(stdouts) + "\n")

    # Log stderrs
    sys.stderr.write('{:-<40}'.format("Out [%s] " % execution_count) + "\n")
    sys.stderr.write("\n".join(stderrs) + "\n")


# Monkey Patch the base preprocess method.
Preprocessor.preprocess = preprocess


def execute_notebook(notebook,
                     output,
                     parameters=None,
                     kernel_name=None,
                     progress_bar=True,
                     log_output=False,
                     start_timeout=60):
    """Executes a single notebook locally.
    Args:
        notebook (str): Path to input notebook.
        output (str): Path to save executed notebook.
        parameters (dict): Arbitrary keyword arguments to pass to the notebook parameters.
        kernel_name (str): Name of kernel to execute the notebook against.
        progress_bar (bool): Flag for whether or not to show the progress bar.
        log_output (bool): Flag for whether or not to write notebook output to stderr.

    Returns:
         nb (NotebookNode): executed notebook object
    """
    print("Input Notebook:  %s" % get_pretty_path(notebook))
    print("Output Notebook: %s" % get_pretty_path(output))
    nb = load_notebook_node(notebook)

    # Parameterize the Notebook.
    if parameters:
        _parameterize_notebook(nb, kernel_name, parameters)

    # Record specified environment variable values.
    nb.metadata.papermill['parameters'] = parameters
    nb.metadata.papermill[
        'environment_variables'] = _fetch_environment_variables()
    nb.metadata.papermill['output_path'] = output

    # Execute the Notebook.
    t0 = datetime.datetime.utcnow()
    processor = ExecutePreprocessor(
        timeout=None,
        startup_timeout=start_timeout,
        kernel_name=kernel_name or nb.metadata.kernelspec.name, )
    processor.progress_bar = progress_bar and not no_tqdm
    processor.log_output = log_output

    processor.preprocess(nb, {})
    t1 = datetime.datetime.utcnow()

    nb.metadata.papermill['start_time'] = t0.isoformat()
    nb.metadata.papermill['end_time'] = t1.isoformat()
    nb.metadata.papermill['duration'] = (t1 - t0).total_seconds()
    nb.metadata.papermill['exception'] = any(
        [cell.metadata.papermill.get('exception') for cell in nb.cells])

    # Write final Notebook to disk.
    write_ipynb(nb, output)
    raise_for_execution_errors(nb, output)
    # always return notebook object
    return nb


def _parameterize_notebook(nb, kernel_name, parameters):

    # Load from a file if 'parameters' is a string.
    if isinstance(parameters, string_types):
        parameters = read_yaml_file(parameters)

    # Generate parameter content based on the kernal_name
    kernel_name = kernel_name or nb.metadata.kernelspec.name
    param_content = _build_parameter_code(kernel_name, parameters)

    param_cell_index = _find_parameters_index(nb)
    if param_cell_index >= 0:
        old_parameters = nb.cells[param_cell_index]
        old_parameters.metadata['tags'].remove('parameters')
        old_parameters.metadata['tags'].append('default parameters')

    newcell = nbformat.v4.new_code_cell(source=param_content)
    newcell.metadata['tags'] = ['parameters']

    before = nb.cells[:param_cell_index + 1]
    after = nb.cells[param_cell_index + 1:]
    nb.cells = before + [newcell] + after


def _build_parameter_code(kernel_name, parameters):
    kernelspec = get_kernel_spec(kernel_name)
    if kernel_name in _parameter_code_builders:
        return _parameter_code_builders[kernel_name](parameters)
    elif kernelspec.language in _parameter_code_builders:
        return _parameter_code_builders[kernelspec.language](parameters)
    raise PapermillException(
        "No parameter builder functions specified for kernel '%s' or language '%s'"
        % (kernel_name, kernelspec.language))


def _find_parameters_index(nb):
    parameters_indices = []
    for idx, cell in enumerate(nb.cells):
        if "parameters" in cell.metadata.tags:
            parameters_indices.append(idx)
    if not parameters_indices:
        return -1
    elif len(parameters_indices) > 1:
        raise PapermillException("Multiple cells with parameters tag found")
    return parameters_indices[0]

def _translate_escaped_str(str_val):
    """Reusable by most interpreters"""
    if isinstance(str_val, string_types):
        str_val = str_val.encode('unicode_escape')
        if sys.version_info >= (3, 0):
            str_val = str_val.decode('utf-8')
        str_val = str_val.replace('"', r'\"')
    return '"{}"'.format(str_val)

def _translate_to_str(val):
    """Reusable by most interpreters"""
    return '{}'.format(val)

# Registry for functions that build parameter assignment code.
_parameter_code_builders = {}

# Python translaters
_translate_type_str_python = _translate_escaped_str
_translate_type_int_python = _translate_to_str
_translate_type_float_python = _translate_to_str
_translate_type_bool_python = _translate_to_str

def _translate_type_dict_python(val):
    escaped = ', '.join(["{}: {}".format(_translate_type_str_python(k), _translate_type_python(v)) for k, v in val.items()])
    return '{{{}}}'.format(escaped)

def _translate_type_list_python(val):
    escaped = ', '.join([_translate_type_python(v) for v in val])
    return '[{}]'.format(escaped)

def _translate_type_python(val):
    """Translate each of the standard json/yaml types to appropiate objects in python."""
    if isinstance(val, string_types):
        return _translate_type_str_python(val)
    # Needs to be before integer checks
    elif isinstance(val, bool):
        return _translate_type_bool_python(val)
    elif isinstance(val, integer_types):
        return _translate_type_int_python(val)
    elif isinstance(val, float):
        return _translate_type_float_python(val)
    elif isinstance(val, dict):
        return _translate_type_dict_python(val)
    elif isinstance(val, list):
        return _translate_type_list_python(val)
    # Use this generic translation as a last resort
    return _translate_escaped_str(val)

# R translaters
_translate_type_str_r = _translate_escaped_str
_translate_type_int_r = _translate_to_str
_translate_type_float_r = _translate_to_str

def _translate_type_bool_r(val):
    return 'TRUE' if val else 'FALSE'

def _translate_type_dict_r(val):
    escaped = ', '.join(["{} = {}".format(_translate_type_str_r(k), _translate_type_r(v)) for k, v in val.items()])
    return 'list({})'.format(escaped)

def _translate_type_list_r(val):
    escaped = ', '.join([_translate_type_r(v) for v in val])
    return 'list({})'.format(escaped)

def _translate_type_r(val):
    """Translate each of the standard json/yaml types to appropiate objects in R."""
    if isinstance(val, string_types):
        return _translate_type_str_r(val)
    # Needs to be before integer checks
    elif isinstance(val, bool):
        return _translate_type_bool_r(val)
    elif isinstance(val, integer_types):
        return _translate_type_int_r(val)
    elif isinstance(val, float):
        return _translate_type_float_r(val)
    elif isinstance(val, dict):
        return _translate_type_dict_r(val)
    elif isinstance(val, list):
        return _translate_type_list_r(val)
    # Use this generic translation as a last resort
    return _translate_escaped_str(val)

def register_param_builder(name):
    """Decorator for registering functions that write variable assignments for a given kernel or language."""

    def wrapper(func):
        _parameter_code_builders[name] = func
        return func

    return wrapper

@register_param_builder("python")
def build_python_params(parameters):
    """Writers parameter assignment code for Python kernels."""
    param_content = "# Parameters\n"
    for var, val in parameters.items():
        param_content += '{} = {}\n'.format(var, _translate_type_python(val))
    return param_content


@register_param_builder("R")
def build_r_params(parameters):
    """Writes parameters assignment code for R kernels."""
    param_content = "# Parameters\n"
    for var, val in parameters.items():
        param_content += '{} = {}\n'.format(var, _translate_type_r(val))
    return param_content


def _fetch_environment_variables():
    ret = dict()
    for name, value in os.environ.items():
        if name in settings.ENVIRONMENT_VARIABLES:
            ret[name] = value
    return ret


ERROR_MESSAGE_TEMPLATE = (
    '<span style="color:red; font-family:Helvetica Neue, Helvetica, Arial, sans-serif; font-size:2em;">'
    "An Exception was encountered at 'In [%s]'."
    '</span>')


def raise_for_execution_errors(nb, output_path):

    error = None
    for cell in nb.cells:
        if cell.get("outputs") is None:
            continue

        for output in cell.outputs:
            if output.output_type == "error":
                error = PapermillExecutionError(
                    exec_count=cell.execution_count,
                    source=cell.source,
                    ename=output.ename,
                    evalue=output.evalue,
                    traceback=output.traceback)
                break

    if error:
        # Write notebook back out with the Error Message at the top of the Notebook.
        error_msg = ERROR_MESSAGE_TEMPLATE % str(error.exec_count)
        error_msg_cell = nbformat.v4.new_markdown_cell(source=error_msg)
        nb.cells = [error_msg_cell] + nb.cells
        write_ipynb(nb, output_path)
        raise error



import unittest
import six
if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch

import pandas as pd
from pandas.util.testing import assert_frame_equal

from .. import display, read_notebook, read_notebooks, PapermillException, record
from ..api import Notebook
from . import get_notebook_path, get_notebook_dir


class TestNotebookClass(unittest.TestCase):

    def test(self):

        path = get_notebook_path('collection/result1.ipynb')
        nb = read_notebook(path)
        self.assertEqual(nb.version, '0.4+2.ge10f94c.dirty')
        self.assertEqual(nb.environment_variables, {})
        self.assertEqual(nb.parameters, dict(foo=1, bar="hello"))
        self.assertEqual(nb.directory, get_notebook_dir('collection/result1.ipynb'))
        expected_df = pd.DataFrame(
            [
                ('bar', 'hello', 'parameter', 'result1.ipynb'),
                ('foo', 1, 'parameter', 'result1.ipynb'),
                ('dict', {u'a': 1, u'b': 2}, 'record', 'result1.ipynb'),
                ('list', [1, 2, 3], 'record', 'result1.ipynb'),
                ('number', 1, 'record', 'result1.ipynb'),
            ],
            columns=['name', 'value', 'type', 'filename']
        )
        assert_frame_equal(nb.dataframe, expected_df)

    def test_bad_file_ext(self):

        with self.assertRaises(PapermillException):
            read_notebook('result_notebook.py')

    def test_path_without_node(self):

        with self.assertRaises(ValueError):
            Notebook(node=None, path='collection/result1.ipynb')


class TestNotebookCollection(unittest.TestCase):

    def test(self):

        path = get_notebook_path('collection')
        nbs = read_notebooks(path)

        expected_df = pd.DataFrame(
            [
                ('bar', 'hello', 'parameter', 'result1.ipynb', 'result1.ipynb'),
                ('foo', 1, 'parameter', 'result1.ipynb', 'result1.ipynb'),
                ('dict', {u'a': 1, u'b': 2}, 'record', 'result1.ipynb', 'result1.ipynb'),
                ('list', [1, 2, 3], 'record', 'result1.ipynb', 'result1.ipynb'),
                ('number', 1, 'record', 'result1.ipynb', 'result1.ipynb'),
                ('bar', 'world', 'parameter', 'result2.ipynb', 'result2.ipynb'),
                ('foo', 2, 'parameter', 'result2.ipynb', 'result2.ipynb'),
                ('dict', {u'a': 1, u'b': 2}, 'record','result2.ipynb', 'result2.ipynb'),
                ('list', [1, 2, 3], 'record', 'result2.ipynb', 'result2.ipynb'),
                ('number', 1, 'record', 'result2.ipynb', 'result2.ipynb'),
            ],
            columns=['name', 'value', 'type', 'filename', 'key']
        )
        assert_frame_equal(nbs.dataframe, expected_df)

        expected_metrics_df = pd.DataFrame(
            [
                ('result1.ipynb', 'Out [1]', 0.0, 'time (s)', 'result1.ipynb'),
                ('result1.ipynb', 'Out [2]', 0.0, 'time (s)', 'result1.ipynb'),
                ('result2.ipynb', 'Out [1]', 0.0, 'time (s)', 'result2.ipynb'),
                ('result2.ipynb', 'Out [2]', 0.0, 'time (s)', 'result2.ipynb'),

            ],
            columns=['filename', 'cell', 'value', 'type', 'key']
        )
        assert_frame_equal(nbs.metrics, expected_metrics_df)


@patch('papermill.api.ip_display')
@patch('IPython.core.formatters.format_display_data')
def test_display(format_display_data_mock, ip_display_mock):
    format_display_data_mock.return_value = ({'foo': 'bar'}, {'metadata': 'baz'})
    display('display_name', {'display_obj': 'hello'})

    format_display_data_mock.assert_called_once_with({'display_obj': 'hello'})
    ip_display_mock.assert_called_once_with({'foo': 'bar'}, metadata={'metadata': 'baz', 'papermill': {'name': 'display_name'}}, raw=True)



@patch('papermill.api.ip_display')
def test_record(ip_display_mock):
    record('a', 3)
    ip_display_mock.assert_called_once_with({'application/papermill.record+json': {'a': 3}}, raw=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Test the command line interface """

import os
import unittest

import pytest
import click
from click.testing import CliRunner

from mock import patch

from .. import cli
from ..cli import papermill, _is_int,  _is_float, _resolve_type




@pytest.mark.parametrize("test_input,expected", [
    ("True", True),
    ("False", False),
    ("None", None),
    ("12.51", 12.51),
    ("10", 10),
    ("hello world", "hello world"),
    (u"😍", u"😍"),
])
def test_resolve_type(test_input, expected):
    assert _resolve_type(test_input) == expected


@pytest.mark.parametrize("value,expected", [
    (13.71, True),
    ("False", False),
    ("None", False),
    (-8.2, True),
    (10, True),
    ("10", True),
    ("12.31", True),
    ("hello world", False),
    (u"😍", False),
])
def test_is_float(value, expected):
    assert (_is_float(value)) == expected


@pytest.mark.parametrize("value,expected", [
    (13.71, True),
    ("False", False),
    ("None", False),
    (-8.2, True),
    ("-23.2", False),
    (10, True),
    ("13", True),
    ("hello world", False),
    (u"😍", False),
])
def test_is_int(value, expected):
    assert (_is_int(value)) == expected


class TestCLI(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.default_args = ['input.ipynb', 'output.ipynb']
        self.sample_yaml_file = os.path.join(os.path.dirname(__file__), 'parameters', 'example.yaml')
        self.sample_json_file = os.path.join(os.path.dirname(__file__), 'parameters', 'example.json')

    @patch(cli.__name__ + '.execute_notebook')
    def test_parameters(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['-p', 'foo', 'bar', '--parameters', 'baz', '42'])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            {'foo': 'bar', 'baz': 42},
            kernel_name=None,
            log_output=False,
            progress_bar=True,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_parameters_raw(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['-r', 'foo', 'bar', '--parameters_raw', 'baz', '42'])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            {'foo': 'bar', 'baz': '42'},
            kernel_name=None,
            log_output=False,
            progress_bar=True,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_parameters_file(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['-f', self.sample_yaml_file, '--parameters_file', self.sample_json_file])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            # Last input wins dict update
            {'foo': 54321, 'bar': 'value', 'baz': {'k2': 'v2', 'k1': 'v1'}},
            kernel_name=None,
            log_output=False,
            progress_bar=True,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_parameters_yaml(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['-y', '{"foo": "bar"}', '--parameters_yaml', '{"foo2": ["baz"]}'])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            {'foo': 'bar', 'foo2': ['baz']},
            kernel_name=None,
            log_output=False,
            progress_bar=True,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_parameters_yaml_override(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['--parameters_yaml', '{"foo": "bar"}', '-y', '{"foo": ["baz"]}'])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            # Last input wins dict update
            {'foo': ['baz']},
            kernel_name=None,
            log_output=False,
            progress_bar=True,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_parameters_base64(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['--parameters_base64', 'eyJmb28iOiAicmVwbGFjZWQiLCAiYmFyIjogMn0=', '-b', 'eydmb28nOiAxfQ=='])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            # Last input wins dict update
            {'foo': 1, 'bar': 2},
            kernel_name=None,
            log_output=False,
            progress_bar=True,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_kernel(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['-k', 'python3'])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            {},
            kernel_name='python3',
            log_output=False,
            progress_bar=True,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_progress_bar(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['--progress-bar'])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            {},
            kernel_name=None,
            log_output=False,
            progress_bar=True,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_no_progress_bar(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['--no-progress-bar'])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            {},
            kernel_name=None,
            log_output=False,
            progress_bar=False,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_log_output(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['--log-output'])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            {},
            kernel_name=None,
            log_output=True,
            progress_bar=True,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_log_output(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['--no-log-output'])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            {},
            kernel_name=None,
            log_output=False,
            progress_bar=True,
            start_timeout=60)

    @patch(cli.__name__ + '.execute_notebook')
    def test_start_timeout(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + ['--start_timeout', '123'])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            {},
            kernel_name=None,
            log_output=False,
            progress_bar=True,
            start_timeout=123)

    @patch(cli.__name__ + '.execute_notebook')
    def test_many_args(self, execute_patch):
        self.runner.invoke(papermill, self.default_args + [
            '-f', self.sample_yaml_file,
            '-y', '{"yaml_foo": {"yaml_bar": "yaml_baz"}}',
            '-b', 'eyJiYXNlNjRfZm9vIjogImJhc2U2NF9iYXIifQ==',
            '-p', 'baz', 'replace',
            '-r', 'foo', '54321',
            '--kernel', 'R',
            '--log-output',
            '--no-progress-bar',
            '--start_timeout', '321',
        ])
        execute_patch.assert_called_with(
            'input.ipynb',
            'output.ipynb',
            {'foo': '54321', 'bar': 'value', 'baz': 'replace', 'yaml_foo': {'yaml_bar': 'yaml_baz'}, "base64_foo": "base64_bar"},
            kernel_name='R',
            log_output=True,
            progress_bar=False,
            start_timeout=321)


def test_cli_path():
    @click.command()
    @click.argument('notebook_path')
    @click.argument('output_path')
    def cli(notebook_path, output_path):
        click.echo('papermill calling %s and %s!' % notebook_path, output_path)

    runner = CliRunner()
    result = runner.invoke(cli,
                           ['papermill', 'notebooks/s3/s3_in/s3-simple_notebook.ipynb',
                            'notebooks/s3/s3_out/output.ipynb'])
    assert result.output


def test_papermill_log():
    @click.group()
    @click.option('--log-output/--no-output', default=False)
    def cli(log_output):
        click.echo('Log output mode is %s' % ('on' if log_output else 'off'))

    @cli.command()
    def papermill():
        click.echo('Logging')

    runner = CliRunner()
    result = runner.invoke(cli, ['--log-output', 'papermill'])
    assert result.exit_code == 0
    assert 'Log output mode is on' in result.output
    assert 'Logging' in result.output



 # -*- coding: utf-8 -*-

from __future__ import unicode_literals

from ansiwrap import strip_color


class AwsError(Exception):
    """Raised when an AWS Exception is encountered."""


class FileExistsError(AwsError):
    """Raised when a File already exists on S3."""


class PapermillException(Exception):
    """Raised when an exception is encountered when operating on a notebook."""


class PapermillExecutionError(PapermillException):
    """Raised when an exception is encountered in a notebook."""

    def __init__(self, exec_count, source, ename, evalue, traceback):
        self.exec_count = exec_count
        self.source = source
        self.ename = ename
        self.evalue = evalue
        self.traceback = traceback
        message = "\n" + 75 * "-" + "\n"
        message += 'Exception encountered at "In [%s]":\n' % str(exec_count)
        message += strip_color("\n".join(traceback))
        message += "\n"

        super(PapermillExecutionError, self).__init__(message)



# -*- coding: utf-8 -*-
"""Utilities for working with S3."""
from __future__ import unicode_literals
from past.builtins import basestring, str

from concurrent import futures
import fnmatch
from functools import wraps, reduce
import gzip
import itertools
import logging
import os
import re
import shutil
import tempfile
import threading
import zlib


import boto3
from boto3.session import Session

import six

from .exceptions import AwsError, FileExistsError


logger = logging.getLogger('papermill.s3')


class Bucket(object):
    """Represents a Bucket of storage on S3

    Parameters
    ----------
    name : string
        name of the bucket
    service : string, optional (Default is None)
        name of a service resource, such as SQS, EC2, etc.

    """

    def __init__(self, name, service=None):
        self.name = name
        self.service = service

    def list(self, prefix='', delimiter=None):
        """Limits a list of Bucket's objects based on prefix and delimiter."""
        return self.service._list(bucket=self.name, prefix=prefix,
                                  delimiter=delimiter, objects=True)


class Prefix(object):
    """Represents a prefix used in an S3 Bucket.

    Parameters
    ----------
    bucket : object
        A bucket of S3 storage
    name : string
        name of the bucket
    service : string, optional (Default is None)
        name of a service resource, such as SQS, EC2, etc.

    """
    def __init__(self, bucket, name, service=None):
        self.bucket = Bucket(bucket, service=service)
        self.name = name
        self.is_prefix = True
        self.service = service

    def __str__(self):
        return 's3://{}/{}'.format(self.bucket.name, self.name)

    def __repr__(self):
        return str(self)


class Key(object):
    """A key that represents a unique object in an S3 Bucket.

    Represents a file or stream.

    Parameters
    ----------
    bucket : object
        A bucket of S3 storage
    name : string
        representative name of the bucket
    size : ???, optional (Default is None)
    etag : ???, optional (Default is None)
    last_modified : date, optional (Default is None)
    storage_class : ???, optional (Default is None)
    service : string, optional (Default is None)
        name of a service resource, such as SQS, EC2, etc.

    """

    # TODO make size, etag, etc properties that can be called from the
    # object as needed
    def __init__(self, bucket, name,
                 size=None, etag=None, last_modified=None, storage_class=None,
                 service=None):
        self.bucket = Bucket(bucket, service=service)
        self.name = name
        self.size = size
        self.etag = etag
        if last_modified:
            try:
                self.last_modified = (
                        last_modified.isoformat().split('+')[0] + '.000Z'
                )
            except ValueError:
                self.last_modified = last_modified
        self.storage_class = storage_class
        self.is_prefix = False
        self.service = service

    def __str__(self):
        return 's3://{}/{}'.format(self.bucket.name, self.name)

    def __repr__(self):
        return str(self)


# retry decorator
def retry(num):
    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for i in range(num):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    logger.debug('retrying: {}'.format(e))
            else:
                raise Exception  # TODO verify what to raise
        return wrapper

    return decorate


class S3(object):
    """Wraps S3.

    Parameters
    ----------
    keyname : TODO
    use_akms : bool, optional (Default is None) (TODO What is akms? Kinesis?)
    host : DEPRECATED. string, optional (Default is None)
    region : string, optional (Default is 'us-east-1')

    Methods
    -------
    The following are wrapped utilities for S3:
        - cat
        - catdir (TODO refactor to cat_dir)
        - cp
        - cpdir (TODO cp_dir)
        - cpmerge (TODO cp_merge)
        - cp_string
        - get_key
        - list
        - list_buckets
        - list_dir
        - list_dir_iterator
        - list_glob
        - list_globs
        - new_folder
        - readdir (TODO read_dir)
        - read
        - rm
        - rmdir (TODO rm_dir)

    """
    sessions = {}
    lock = threading.RLock()

    # TODO: add support for assume role.

    def __init__(self,
                 keyname=None, use_akms=None, host=None,
                 region='us-east-1', *args, **kwargs):

        import botocore.session

        # DEPRECATED - `host`
        if host:
            logger.warning('the host param is deprecated')

        use_akms = (use_akms if use_akms is not None
                    else 'USE_AKMS' in os.environ)

        args = (keyname, use_akms)
        with self.lock:
            if args not in self.sessions:
                session = Session(
                    botocore_session=botocore.session.Session(
                        session_vars={
                            'akms_keyname': keyname,
                            'use_akms': use_akms,
                        }
                    )
                )
                client = session.client('s3')
                s3 = session.resource('s3')
                self.sessions[args] = (session, client, s3)

        (self.session, self.client, self.s3) = self.sessions[args]

    def __batches(self, it, size=1000):
        while True:
            data = itertools.islice(it, size)
            batch = list(data)
            if not batch:
                break
            yield batch

    def __create_callback(self, filename, total=None, callback=None):
        if callback is None:
            cur = 0  # TODO refactor cur name

            class Callback(object):
                def __init__(self, total):
                    self.cur = 0
                    self.total = total

                def __call__(self, retr):

                    if self.total:
                        self.cur += retr  # TODO refactor cur and retr names
                        logger.debug(
                            '%s: %6d/%dkb (%3.2f%% Complete)' % filename,
                            self.cur / 1024, self.total / 1024,
                            self.cur / float(total) * 100
                        )

        return callback

    def _bucket(self, bucket):
        if isinstance(bucket, Key):
            return bucket.bucket
        return Bucket(self._bucket_name(bucket), service=self)

    def _bucket_name(self, bucket):
        return self._clean(bucket).split('/', 1)[0]

    def _clean(self, name):
        if name.startswith('s3n:'):
            name = 's3:' + name[4:]
        if self._is_s3(name):
            return name[5:]
        return name

    def _clean_s3(self, name):
        return 's3:' + name[4:] if name.startswith('s3n:') else name

    @retry(3)
    def _copy(self, source, dest):
        key = self._get_key(dest)
        source_key = self._get_key(source)
        obj = self.s3.Object(key.bucket.name, key.name)
        obj.copy_from(
            CopySource='{}/{}'.format(source_key.bucket.name, source_key.name)
        )
        return key

    @retry(3)
    def _get(self, source, dest=None, num_callbacks=10, **kwargs):
        assert dest is not None, 'a destination must be provided'

        targetFile = isinstance(dest, six.string_types)
        if targetFile:
            tmpfile = tempfile.NamedTemporaryFile()
        else:
            tmpfile = dest

        try:
            key = self._get_key(source)
            obj = self.s3.Object(key.bucket.name, key.name)
            obj.load()
            obj.download_fileobj(
                tmpfile,
                Callback=self.__create_callback(
                    source,
                    total=obj.content_length,
                    callback=kwargs.get('callback')
                )
            )
            if targetFile:
                tmpfile.seek(0)
                shutil.copyfileobj(tmpfile, open(dest, 'wb'))
            return key
        finally:
            if targetFile:
                tmpfile.close()

    def _get_key(self, name):
        if isinstance(name, Key):
            return name

        return Key(bucket=self._bucket_name(name),
                   name=self._key_name(name),
                   service=self)

    def _get_file_obj(self, f, mode='wb'):
        return gzip.open(f, mode) if f.endswith('.gz') else open(f, mode)

    def _key_name(self, name):
        cleaned = self._clean(name).split('/', 1)
        return cleaned[1] if len(cleaned) > 1 else None

    @retry(3)
    def _list(self,
              prefix='',
              bucket=None,
              delimiter=None,
              keys=False,
              objects=False,
              page_size=1000,
              **kwargs):
        assert bucket is not None, 'You must specify a bucket to list'

        bucket = self._bucket_name(bucket)
        paginator = self.client.get_paginator('list_objects_v2')
        operation_parameters = {
            'Bucket': bucket,
            'Prefix': prefix,
            'PaginationConfig': {'PageSize': page_size},
        }
        if delimiter:
            operation_parameters['Delimiter'] = delimiter

        page_iterator = paginator.paginate(**operation_parameters)

        def sort(item):
            if 'Key' in item:
                return item['Key']
            return item['Prefix']

        for page in page_iterator:
            locations = (
                sorted(
                    [i for i in page.get('Contents', []) + page.get('CommonPrefixes', [])],
                    key=sort
                )
            )

            for item in locations:
                if objects or keys:
                    if 'Key' in item:
                        yield Key(
                            bucket,
                            item['Key'],
                            size=item.get('Size'),
                            etag=item.get('ETag'),
                            last_modified=item.get('LastModified'),
                            storage_class=item.get('StorageClass'),
                            service=self
                        )
                    elif objects:
                        yield Prefix(bucket, item['Prefix'], service=self)
                else:
                    prefix = item['Key'] if 'Key' in item else item['Prefix']
                    yield 's3://{}/{}'.format(bucket, prefix)

    def _put(self, source, dest, num_callbacks=10,
             policy='bucket-owner-full-control', **kwargs):
        key = self._get_key(dest)
        obj = self.s3.Object(key.bucket.name, key.name)
        length = 0

        # support passing in open file obj.  Why did we do this in the past?

        if not isinstance(source, six.string_types):
            obj.upload_fileobj(
                source,
                ExtraArgs={'ACL': policy},
                Callback=self.__create_callback(
                    dest,
                    total=os.fstat(source.fileno()).st_size,
                    callback=kwargs.get('callback')
                )
            )
        else:
            obj.upload_file(
                source,
                ExtraArgs={'ACL': policy},
                Callback=self.__create_callback(
                    dest,
                    total=length,
                    callback=kwargs.get('callback')
                )
            )
        return key

    def _put_string(self, source, dest, num_callbacks=10,
                    policy='bucket-owner-full-control', **kwargs):
        key = self._get_key(dest)
        obj = self.s3.Object(key.bucket.name, key.name)
        length = len(source)

        if isinstance(source, str):
            source = source.encode('utf-8')
        obj.put(Body=source, ACL=policy)
        return key

    def _is_s3(self, name):
        # only allow file objects from local
        if not isinstance(name, (basestring, Key, Prefix)):
            return False

        name = self._clean_s3(name)
        return 's3://' in name

    def _sizeof(self, key, bucket=None):
        assert bucket is not None, 'You must specify a bucket'

        response = self.client.head_object(Bucket=bucket, Key=key)
        return response['ContentLength']

    def cat(self, source, buffersize=None, memsize=2**24, compressed=False,
            encoding='UTF-8', raw=False):
        """Returns an iterator for the data in the key or nothing if the key
        doesn't exist. Decompresses data on the fly (if compressed is True
        or key ends with .gz) unless raw is True. Pass None for encoding to
        skip encoding.
        """
        assert (self._is_s3(source) or isinstance(source, Key)), \
                'source must be a valid s3 path'

        key = self._get_key(source) if not isinstance(source, Key) else source
        compressed = (compressed or key.name.endswith('.gz')) and not raw
        if compressed:
            decompress = zlib.decompressobj(16 + zlib.MAX_WBITS)

        size = 0
        bytes_read = 0
        err = None
        undecoded = ''
        if key:
            # try to read the file multiple times
            for i in range(100):
                obj = self.s3.Object(key.bucket.name, key.name)
                buffersize = buffersize if buffersize is not None else 2**20

                if not size:
                    size = obj.content_length
                elif size != obj.content_length:
                    raise AwsError(
                        'key size unexpectedly changed while reading')

                r = obj.get(Range="bytes={}-".format(bytes_read))

                try:
                    while bytes_read < size:
                        # this making this weird check because this call is
                        # about 100 times slower if the amt is too high
                        if size - bytes_read > buffersize:
                            bytes = r['Body'].read(amt=buffersize)
                        else:
                            bytes = r['Body'].read()
                        if compressed:
                            s = decompress.decompress(bytes)
                        else:
                            s = bytes

                        if encoding and not raw:
                            try:
                                decoded = undecoded + s.decode(encoding)
                                undecoded = ''
                                yield decoded
                            except UnicodeDecodeError as ude:
                                undecoded += s
                                if len(undecoded) > memsize:
                                    raise
                        else:
                            yield s

                        bytes_read += len(bytes)

                except zlib.error as e:
                    logger.error("Error while decompressing [%s]", key.name)
                    raise
                except UnicodeDecodeError as ude:
                    raise
                except Exception as e:
                    err = True
                    pass

                if size <= bytes_read:
                    break

            if size != bytes_read:
                if err:
                    raise Exception
                else:
                    raise AwsError('Failed to fully read [%s]' % source.name)

            if undecoded:
                assert encoding is not None  # only time undecoded is set

                # allow exception to be raised if one is thrown
                decoded = undecoded.decode(encoding)
                yield decoded

    def catdir(self, source, buffersize=None, compressed=False,
               encoding='UTF-8'):
        """Iterates over a dir in s3 split on newline.

        Yields chunks of data from the file.

        """
        for f in self.listdir(source, keys=True):
            logger.debug('S3.catdir: %s', f)
            full_name = 's3://{}/{}'.format(f.bucket.name, f.name)
            for l in self.cat(full_name, buffersize=buffersize,
                              compressed=compressed, encoding=encoding):
                yield l

    def cp(self, source, dest, **kwargs):
        """Copies to and from an s3 bucket to a file.

        The source and destination should be the fully qualified path of an
        s3 key and the filename to copy from/to.

        """
        fr = self._is_s3(source)
        to = self._is_s3(dest)
        if fr and to:
            return self._copy(source, dest, **kwargs)
        elif fr:
            return self._get(source, dest, **kwargs)
        else:
            return self._put(source, dest, **kwargs)

    def cpdir(self, source, dest, recursive=True, **kwargs):
        """Copies to and from an s3 prefix to a directory.  The source and
        destination should be the fully qualified path of an s3 prefix and the
        directory to copy from/to.

        """
        source = source.endswith('/') and source or "%s/" % source
        source = self._clean_s3(source)
        dest = re.sub('/$', '', dest)

        fr = self._is_s3(source)
        to = self._is_s3(dest)

        if fr:
            for i in list(self.listdir(source)):
                file = "%s/%s" % (dest, i.replace(source, ''))
                if not i.endswith('/'):
                    self.cp(i, file, **kwargs)
                elif recursive:
                    if not to and not os.path.exists(file):
                        os.makedirs(file)
                    self.cpdir(i, file, recursive=recursive, **kwargs)
        else:
            for i in os.listdir(source):
                s = "%s/%s" % (source, i)
                d = "%s/%s" % (dest, i)
                if recursive and os.path.isdir(s):
                    self.cpdir(s, d, recursive=recursive, **kwargs)
                else:
                    self._put(s, d, **kwargs)

    def cpmerge(self, source, dest):
        """Takes a source directory and a destination file as input and
        concatenates files in src into the destination local file.
        """
        fr = self._is_s3(source)
        to = self._is_s3(dest)

        if fr and not to:
            if os.path.isfile(dest):
                raise FileExistsError("File {} already exists".format(dest))
            if os.path.isdir(dest):
                raise AwsError("Destination {} is a directory".format(dest))

            try:
                os.makedirs(os.path.dirname(dest))
            except (OSError) as e:
                if e.strerror != 'File exists':
                    raise

            with self._get_file_obj(dest) as outfile:
                for f in self.listdir(source):
                    logger.debug('reading {}'.format(f))
                    for line in self.cat(f, compressed=f.endswith('gz')):
                        outfile.write(line)
        else:
            raise AwsError(
                "Copy-merging {} to {} is not supported".format(source, dest))

    def cp_string(self, source, dest, **kwargs):
        """
        Copies source string into the destination location.

        Parameters:
            source: the string with the content to copy
            dest: the s3 location

        Uses basestring type (python2) when checking if a string

        """

        assert isinstance(source, basestring), "source must be a string"
        assert self._is_s3(dest), "Destination must be s3 location"

        return self._put_string(source, dest, **kwargs)

    def get_key(self, name):
        """
        Get Key object if the name represents a key, otherwise return None.

        Examples:

        >>> s3.get_key('s3://netflix-dataoven-test-users/jwalton/nothing.py')

            <Key: netflix-dataoven-test-users,jwalton/nothing.py>

        >>> s3.get_key('s3://netflix-dataoven-test-users/jwalton/')

            None

        Args:
           name (str): the full s3 location for which to get the key

        Returns:
           A boto3.resources.factory.s3.Object object if the name represents
           a file or None if the name is a prefix

        """
        assert self._is_s3(name), "location must be in form s3://bucket/key"

        key = self._get_key(name)
        try:
            obj = self.s3.Object(key.bucket.name, key.name)
            obj.load()
            return obj
        except boto3.exceptions.botocore.exceptions.ClientError as e:
            logger.debug(e)
            pass

    def list(self, name, iterator=False, **kwargs):
        """ Returns a list of the files under the specified path
        name must be in the form of s3://bucket/prefix

        optional params
        ---------------
        keys: if True then this will return the actual boto keys for files
              that are encountered
        objects: if True then this will return the actual boto objects for
                 files or prefixes that are encountered
        delimiter: if set this
        iterator: if True return iterator rather than converting to list
                  object

        """
        assert self._is_s3(name), "name must be in form s3://bucket/key"

        it = self._list(bucket=self._bucket_name(name),
                        prefix=self._key_name(name),
                        **kwargs)
        return iter(it) if iterator else list(it)

    def list_buckets(self):
        return [
            bucket['Name']
            for bucket in self.client.list_buckets().get('Buckets', [])
        ]

    def listdir(self, name, **kwargs):
        """ Returns a list of the files under the specified path.

        This is different from list as it will only give you files under the
        current directory, much like ls.

        name must be in the form of s3://bucket/prefix/

        optional params:
        keys: if True then this will return the actual boto keys for files
            that are encountered
        objects: if True then this will return the actual boto objects for
            files or prefixes that are encountered
        """
        assert self._is_s3(name), "name must be in form s3://bucket/prefix/"

        if not name.endswith('/'):
            name += "/"
        return self.list(name, delimiter='/', **kwargs)

    def listdir_iterator(self, name):
        """ Iterates over a list of objects in a bucket.

        Calls `bucket.list` which returns a
        `boto.s3.bucketlistresultset.bucketlistresultset`

        Items returned from iterator are `boto` `objects`.

        """
        assert self._is_s3(name), "name must be in form s3://bucket/key"

        return self._list(
            prefix=self._key_name(name),
            bucket=self._bucket_name(name),
            objects=True,
            delimiter='/')

    def listglob(self, glb, **kwargs):
        """ Returns a list of the files matching a glob.

        Name must be in the form of s3://bucket/glob

        """
        r = []
        regex = re.compile('[*?\[]')
        for file in self.list(regex.split(glb, 1)[0], **kwargs):
            if fnmatch.fnmatch(str(file), glb):
                r.append(file)
        return r

    def listglobs(self, *args, **kwargs):
        """ Returns a list of files for multiple glob operators.

        Equivalent to list(listglob(args[0])) + list(listglob(args[1]))  + ...

        Use Python lambda and reduce

        arguments acc and x
        sequence acc + x
        executor.map(self.listglob, args)
        []

        The function reduce(func, seq) continually applies the function
        func() to the sequence seq. It returns a single value.


        """
        with futures.ThreadPoolExecutor(max_workers=kwargs.get('threads', len(args))) as executor:
            out = reduce(lambda acc, x: acc + x, executor.map(self.listglob, args), [])
        return out

    def new_folder(self, name):
        """
        Creates a new "folder" by adding a file called "_empty" under the prefix given.

        Example:
            >>> s3.new_folder('s3://thebucket/somedir/newdir/')

        Args:
            name - the full location for the new "folder" in the format s3://bucket/prefix

        Returns: key for the new "folder"

        """
        assert self._is_s3(name), "name must be in form s3://bucket/new_key/"

        if not name.endswith("/"):
            name += "/"

        existing = self.get_key(name)
        if existing is not None:
            raise Exception("key {} already exists".format(name))

        loc = name + "_empty"
        self._put_string("", loc)

        return self.get_key(name)

    def readdir(self, source, compressed=False, encoding='UTF-8'):
        """Iterates over a dir in s3 split on newline.

        Yields line in file in dir.

        """
        for f in self.listdir(source, keys=True):
            for l in self.read(str(f),
                               compressed=compressed,
                               encoding=encoding):
                yield l

    def read(self, source, compressed=False, encoding='UTF-8'):
        """Iterates over a file in s3 split on newline.

        Yields a line in file.

        """
        buf = ''
        for block in self.cat(source,
                              compressed=compressed,
                              encoding=encoding):
            buf += block
            if '\n' in buf:
                ret, buf = buf.rsplit('\n', 1)
                for line in ret.split('\n'):
                    yield line

        lines = buf.split('\n')
        for line in lines[:-1]:
            yield line

        # only yield the last line if the line has content in it
        if lines[-1]:
            yield lines[-1]

    def rm(self, target):
        """Remove a single key from s3.

        If you want to remove a directory use rmdir.

        """
        assert self._is_s3(target), 'target must be a valid s3 path'

        key = self._get_key(target)
        if key:
            obj = self.s3.Object(key.bucket.name, key.name)
            return obj.delete()

    def rmdir(self, target, delimiter='/'):
        """Removes every key that falls under the target directory."""
        assert self._is_s3(target), 'target must be a valid s3 path'

        bucket = self.s3.Bucket(self._bucket_name(target))
        errors = []
        for batch in self.__batches(
                self._list(
                    bucket=self._bucket_name(target),
                    prefix=self._key_name(target),
                    keys=True)):
            keys = [dict(Key=n.name) for n in batch]
            response = bucket.delete_objects(Delete={'Objects': keys})
            if 'Errors' in response:
                errors += response['Errors']
        if errors:
            logger.warning("errors deleting: {}".format(errors))


# leaving this here for compatibility
def split(path):
    """
    Splits an s3 path into bucket and prefix, like `os.path.split`.

    It can only be used once.

    For example, `s3://foo/bar/baz` would return `['foo','bar/baz']`

    .. note::
       A trailing `/` is significant in s3, and it will not be stripped,
       ie `s3://foo/bar/baz/` will return `['foo','bar/baz/']`

    """
    if not path.startswith('s3://'):
        raise ValueError('path must start with s3://')
    path = path[5:]
    if path:
        if '/' in path:
            return path.split('/', 1)
        else:
            return [path, '']
    else:
        return ['', '']




#!/usr/bin/env python
# -*- coding: utf-8 -*-
""""
setup.py

Note: Install ipython version based on python version.
      If python 2, install ipython 5.x (python2 is not supported in ipython 6.

"""
import os
import sys
from os.path import exists
from setuptools import setup

import versioneer

ipython_req = 'ipython'

python_2 = sys.version_info[0] == 2
def read(fname):
    with open(fname, 'rU' if python_2 else 'r') as fhandle:
        return fhandle.read()

req_path = os.path.join(os.path.dirname('__file__'), 'requirements.txt')
required = [req.strip() for req in read(req_path).splitlines() if req.strip()]
if python_2 and 'bdist_wheel' not in sys.argv:
    required = ['ipython<6' if req == 'ipython' else req for req in required]

setup(name='papermill',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Map Reduce for Notebooks',
      author='nteract contributors',
      author_email='jupyter@googlegroups.com',
      license='BSD',
      keywords="jupyter mapreduce",
      long_description=(open('README.rst').read() if exists('README.rst') else ''),
      url='https://github.com/nteract/papermill',
      packages=['papermill'],
      install_requires=required,
      entry_points={
              'console_scripts': [
                  'papermill = papermill.cli:papermill'
              ]
          }
      )



import unittest

import six

from ..api import read_notebook
from ..execute import _parameterize_notebook
from . import get_notebook_path

PYTHON = 'python2' if six.PY2 else 'python3'


class TestNotebookParametrizing(unittest.TestCase):
    def test_not_preserving_tags(self):
        # test that other tags on the parameter cell are _not_ preserved
        test_nb = read_notebook(get_notebook_path("simple_execute.ipynb"))
        test_nb.node.cells[0]['metadata']['tags'].append('some tag')

        _parameterize_notebook(test_nb.node, PYTHON, {'msg': 'Hello'})

        cell_zero = test_nb.node.cells[0]
        self.assertTrue('some tag' in cell_zero.get('metadata').get('tags'))

        cell_one = test_nb.node.cells[1]
        self.assertTrue('some tag' not in cell_one.get('metadata').get('tags'))
        self.assertTrue('parameters' in cell_one.get('metadata').get('tags'))

    def test_default_parameters_tag(self):
        test_nb = read_notebook(get_notebook_path("simple_execute.ipynb"))

        _parameterize_notebook(test_nb.node, PYTHON, {'msg': 'Hello'})

        cell_zero = test_nb.node.cells[0]
        self.assertTrue('default parameters'
                        in cell_zero.get('metadata').get('tags'))
        self.assertTrue('parameters'
                        not in cell_zero.get('metadata').get('tags'))

    def test_no_parameters(self):
        test_nb = read_notebook(get_notebook_path("simple_execute.ipynb"))
        test_nb.node.cells[0]['metadata']['tags'].append('some tag')

        _parameterize_notebook(test_nb.node, PYTHON, {'msg': 'Hello'})

        cell_zero = test_nb.node.cells[0]
        self.assertTrue('default parameters'
                        in cell_zero.get('metadata').get('tags'))
        self.assertTrue('parameters'
                        not in cell_zero.get('metadata').get('tags'))



# -*- coding: utf-8 -*-
"""Main `papermill` interface."""
from __future__ import unicode_literals

import base64

import click
import yaml

from .execute import execute_notebook
from .iorw import read_yaml_file


@click.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.argument('notebook_path')
@click.argument('output_path')
@click.option('--parameters', '-p', nargs=2, multiple=True,
              help='Parameters to pass to the parameters cell.')
@click.option('--parameters_raw', '-r', nargs=2, multiple=True,
              help='Parameters to be read as raw string.')
@click.option('--parameters_file', '-f', multiple=True,
              help='Path to YAML file containing parameters.')
@click.option('--parameters_yaml', '-y', multiple=True,
              help='YAML string to be used as parameters.')
@click.option('--parameters_base64', '-b', multiple=True,
              help='Base64 encoded YAML string as parameters.')
@click.option('--kernel', '-k', help='Name of kernel to run.')
@click.option('--progress-bar/--no-progress-bar', default=True,
              help="Flag for turning on the progress bar.")
@click.option('--log-output/--no-log-output', default=False,
              help="Flag for writing notebook output to stderr.")
@click.option('--start_timeout', type=int, default=60,
              help="Time in seconds to wait for kernel to start.")
def papermill(notebook_path, output_path, parameters, parameters_raw,
              parameters_file, parameters_yaml, parameters_base64, kernel,
              progress_bar, log_output, start_timeout):
    """This utility executes a single notebook on a container.

    Papermill takes a source notebook, applies parameters to the source
    notebook, executes the notebook with the specified kernel, and saves the
    output in the destination notebook.

    """
    # Read in Parameters
    parameters_final = {}
    for params in parameters_base64 or []:
        parameters_final.update(yaml.load(base64.b64decode(params)))
    for files in parameters_file or []:
        parameters_final.update(read_yaml_file(files))
    for params in parameters_yaml or []:
        parameters_final.update(yaml.load(params))
    for name, value in parameters or []:
        parameters_final[name] = _resolve_type(value)
    for name, value in parameters_raw or []:
        parameters_final[name] = value

    execute_notebook(notebook_path, output_path, parameters_final,
                     kernel_name=kernel, progress_bar=progress_bar,
                     log_output=log_output, start_timeout=start_timeout)


def _resolve_type(value):
    if value == "True":
        return True
    elif value == "False":
        return False
    elif value == "None":
        return None
    elif _is_int(value):
        return int(value)
    elif _is_float(value):
        return float(value)
    else:
        return value


def _is_int(value):
    """Use casting to check if value can convert to an `int`."""
    try:
        int(value)
    except ValueError:
        return False
    else:
        return True


def _is_float(value):
    """Use casting to check if value can convert to a `float`."""
    try:
        float(value)
    except ValueError:
        return False
    else:
        return True


if __name__ == '__main__':
    papermill()
